/* Mastermind-Bibliothek.
 *
 * Stefan Arnold XI.03
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#include "Mastermind.h"


/* Liefert eine Bewertung des angegebenen Tips bei gegebenem Code
 * nach den auf Uebungsblatt 5 spezifizierten Regeln.
 */
bewertung getBewertung(code Code, code Tip, zahl Optionen)
{
    /* lokale Kopien, da Veraenderungen vorgenommen werden */
    code C,T;
    const zahl GELESEN=Optionen;

    bewertung result;
    int i,j;

    result.schwarz=0;
    result.weiss=0;
    for(i=0;i<CODELAENGE;i++)
    {
        C[i]=Code[i];
        T[i]=Tip[i];
    }

    /* schwarze Stifte vergeben */
    for(i=0;i<CODELAENGE;i++)
    {
        if (C[i]==T[i])
        {
            result.schwarz++;
            C[i]=GELESEN;
            T[i]=GELESEN;
        }
    }

    /* weisse Stifte vergeben */
    for(i=0;i<CODELAENGE;i++)
    {
        if(C[i]==GELESEN)
            continue;
        for(j=0;j<CODELAENGE;j++)
        {
            if(T[j]==GELESEN)
                continue;
            if (C[i]==T[j])
            {
                result.weiss++;
                T[j]=GELESEN;
                break;
            }
        }
    }
    return result;
}


/* Gibt die im angegebenen Code enthaltenen Zahlen aus.
 */
void printCode(code Code)
{
    int i;
    printf("%u",Code[0]);
    for (i=1;i<CODELAENGE;i++)
        printf("; %u",Code[i]);
}

/* Gibt die angegebene Bewertung aus.
 */
void printBewertung(bewertung B)
{
    printf("schwarze Stifte: %u; weisse Stifte: %u",B.schwarz,B.weiss);
}

/* Liest Daten vom Typ 'code' ein. Die Zahlen muessen dabei
 * im Bereich 0..'maxZahl' liegen. Liefert...
 *   0, falls das Einlesen erfolgreich war,
 *  das Ergebnis von 'scanf',
 *      wenn ein unbekannter Fehler
 *      aufgetreten ist,
 *  1, falls eine Zahl groesser als 'maxZahl' war
 *      oder nicht genuegend Zahlen eingelesen wurden,
 *      oder eine Zahl doppelt eingegeben wurde, obwohl
 *      ohne Vielfachheiten gespielt wird.
 *   .
 */
int readCode(code Code,zahl Optionen,int mitVielfachheiten)
{
    int i,j,check;

    printf("Ihr Tip:\n");
    fflush(stdin);
    for(i=0;i<CODELAENGE;i++)
    {
        check=scanf("%u",&Code[i]);
        if(check==0)
            return 1;
        if(check!=1)
            return check;
        if(Code[i]>=Optionen)
            return 1;
    }
    if (!mitVielfachheiten)
    {
        for(i=0;i<CODELAENGE;i++)
            for(j=0;j<i;j++)
                if(Code[j]==Code[i])
                    return 1;
    }
    return 0;
}

/* Liest Daten vom Typ 'bewertung' ein. Die Summe von
 * 'schwarz' und 'weiss' muss dabei im Bereich 0..CODELAENGE
 * liegen. Weitere Kontrollen finden nicht statt. Liefert...
 *   0, falls das Einlesen erfolgreich war,
 *  das Ergebnis von 'scanf',
 *      wenn ein unbekannter Fehler
 *      aufgetreten ist,
 *  1, falls die Summe von 'schwarz' und 'weiss' ausserhalb
 *      des gueltigen Bereichs liegt,
 *      oder nicht genuegend Zahlen eingelesen wurden.
 *   .
 */
int readBewertung(bewertung* b)
{
    int check,s,w;

    printf("Ihre Bewertung:\n");
    printf("schwarz: ");
    fflush(stdin);
    check=scanf("%u",&s);
    if(check==0)
        return 1;
    if(check!=1)
        return check;
    printf("weiss: ");
    fflush(stdin);
    check=scanf("%u",&w);
    if(check==0)
        return 1;
    if(check!=1)
        return check;

    if (s+w<=CODELAENGE)
    {
        (*b).schwarz=(zahl)s;
        (*b).weiss=(zahl)w;
        return 0;
    }
    else
    {
        (*b).schwarz=CODELAENGE+1;
        (*b).weiss=CODELAENGE+1;
        return 1;
    }
return 0;
}

/* Berechnet die Wertigkeiten, die zum Extrahieren von Codes benoetigt
 * werden.
 */
void berechneWertigkeiten(wertigkeit W,zahl Optionen)
{
    int i;

    if (Optionen<CODELAENGE)
    {
        for(i=0;i<CODELAENGE;i++)
            W[i]=1;
        return;
    }

    for(i=0;i<CODELAENGE;i++)
        W[i]=(zahl)(Optionen-i-1);

    W[CODELAENGE-1]=1;

    for(i=CODELAENGE-2;i>=0;i--)
        W[i]*=W[i+1];
}

/* Berechnet den zum angegebenen Index gehoerenden Code.
 */
void extrahiereCode(index src,code dst,wertigkeit W,zahl Optionen)
{
    zahl* Zahlen;
    int i,j;

    /* 'Zahlen' soll alle moeglichen Zahlen enthalten. */
    Zahlen=calloc(Optionen,sizeof(zahl));
    for(i=0;i<Optionen;i++)
        Zahlen[i]=(zahl)i;

    for(i=0;i<CODELAENGE;i++)
    {
        j=(src/W[i]);
        dst[i]=Zahlen[j];
        for (;j<Optionen-1-i;j++)
            Zahlen[j]=Zahlen[j+1];
        src%=W[i];
    }
    free(Zahlen);
}

/* Erzeugt einen zufaelligen Code.
 */
void ZufallsCode(code Code,zahl Optionen,int Vielfachheiten)
{
    int i,j;
    srand((unsigned)time(NULL));

    if (Vielfachheiten)
    {
        for(i=0;i<CODELAENGE;i++)
            Code[i]=(zahl)(rand()%Optionen);
    }
    else
    {
        for(i=0;i<CODELAENGE;i++)
        {
            Code[i]=(zahl)(rand()%Optionen);
            for(j=0;j<i;j++)
                if (Code[j]==Code[i])
                {
                    i--;
                    break;
                }
        }
    }
}

/* Erwartungswert der verbleibenden noetigen Tips, wenn der
 * angegebene Tip gewaehlt wird.
 */
int Erwartungswert(code Tip,menge M,wertigkeit W,zahl Optionen)
{
    int i,j,k=0,*sum,result=-1;
    index l;
    bewertung *B,b;
    code Code;

    /* k: Anzahl theoretisch moeglicher Bewertungen */
    for(i=0;i<=CODELAENGE;i++)
        for(j=0;j<=CODELAENGE-i;j++)
            k++;
    if(k==0)
        return -1;

    /* B soll alle theoretisch moeglichen Bewertungen enthalten. */
    B=calloc(k,sizeof(bewertung));
    for(i=k=0;i<=CODELAENGE;i++)
    {
        for(j=0;j<=CODELAENGE-i;j++)
        {
            B[k].schwarz=(zahl)i;
            B[k].weiss=(zahl)j;
            k++;
        }
    }

    sum=calloc(k,sizeof(int));
    for(i=0;i<k;i++)
        sum[i]=0;

    /* Iteration ueber alle Codes */
    for(l=0;l<M.size;l++)
    {
        /* Ist der Code noch moeglich? */
        if (istElement(l,M))
        {
            extrahiereCode(l,Code,W,Optionen);
            b=getBewertung(Code,Tip,Optionen);
            /* Einordnung bei der entsprechenden Bewertung */
            for(i=0;i<k;i++)
            {
                if (b.schwarz==B[i].schwarz && b.weiss==B[i].weiss)
                {
                    if(result<0)
                        result=0;
                    sum[i]++;
                    break;
                }
            }
        }
    }
    for(i=0;i<k;i++)
        result+=sum[i]*sum[i];

    free(sum);
    free(B);
    return result;
}

/* Berechnet einen der Tips, fuer die 'Erwartungswert' ein
 * (nichtnegatives) Minimum annimmt und schreibt diesen in
 * 'dst'.
 * Liefert den zugehoerigen Erwartungswert verbleibender Tips
 * oder -1, falls kein Tip mehr moeglich ist.
 */
int berechneTip(code dst,menge M,zahl Optionen)
{
    int i,j,n=0,EW,minEW=-1;
    index k=0;
    bewertung* B;
    code tmpTip;
    wertigkeit W;

    berechneWertigkeiten(W,Optionen);

    /* n: Anzahl theoretisch moeglicher Bewertungen */
    for(i=0;i<CODELAENGE;i++)
        for(j=0;j<CODELAENGE-i;j++)
            n++;

    /* B soll alle theoretisch moeglichen Bewertungen enthalten. */
    B=calloc(n,sizeof(bewertung));
    for(i=k=0;i<CODELAENGE;i++)
        for(j=0;j<CODELAENGE-i;j++)
        {
            B[k].schwarz=(zahl)i;
            B[k].weiss=(zahl)j;
            k++;
        }

    for(k=0;k<M.size;k++)
    {
        if (istElement(k,M))
        {
            extrahiereCode(k,tmpTip,W,Optionen);
            EW=Erwartungswert(tmpTip,M,W,Optionen);
            if (EW>=0 && (EW<minEW || minEW<0))
            {
                minEW=EW;
                for(i=0;i<CODELAENGE;i++)
                    dst[i]=tmpTip[i];
            }
        }
    }
    free(B);
    return minEW;
}

/* Entfernt alle Elemente aus M, die aufgrund des angegebenen
 * Tips und der Bewertung keine moeglichen Loesungen mehr sind.
 */
void berechneVerbleibende(menge M,code Tip,bewertung b,zahl Optionen)
{
    index i;
    int j,k,schwarz,weiss;
    code tmp;
    wertigkeit W;

    berechneWertigkeiten(W,Optionen);
    for(i=0;i<M.size;i++)
    {
        if (istElement(i,M))
        {
            extrahiereCode(i,tmp,W,Optionen);

            schwarz=weiss=0;

            for(j=0;j<CODELAENGE;j++)
            {
                if (tmp[j]==Tip[j])
                {
                    schwarz++;
                    /* alle richtigen im folgenden ignorieren: */
                    tmp[j]=Optionen;
                    /* naechstes Zeichen suchen */
                    continue;
                }
                for(k=0;k<CODELAENGE;k++)
                {
                    if (tmp[k]==Tip[j])
                    {
                        weiss++;
                        /* alle richtigen im folgenden ignorieren: */
                        tmp[k]=Optionen;
                        /* naechstes Zeichen suchen */
                        break;
                    }
                }
            }
            if(schwarz!=b.schwarz || weiss!=b.weiss)
                /* Element entfernen */
                invertiereElement(i,M);
        }
    }
}

/* Liefert genau dann 0, wenn i kein Element von M ist.
 */
__inline int istElement(index i,menge M)
{
    unsigned short Maske;

    switch (i%8)
    {
        case 0:  Maske=1;   break;
        case 1:  Maske=2;   break;
        case 2:  Maske=4;   break;
        case 3:  Maske=8;   break;
        case 4:  Maske=16;  break;
        case 5:  Maske=32;  break;
        case 6:  Maske=64;  break;
        case 7:  Maske=128; break;
        /* dieser Fall sollte eigentlich nie eintreten: */
        default: Maske=0;   break;
    }
    return Maske & M.elem[i/8];
}

/* Ist i Element von M, wird es entfernt, sonst hinzugefuegt.
 */
__inline void invertiereElement(index i,menge M)
{
    unsigned short Maske;

    switch (i%8)
    {
        case 0:  Maske=1;   break;
        case 1:  Maske=2;   break;
        case 2:  Maske=4;   break;
        case 3:  Maske=8;   break;
        case 4:  Maske=16;  break;
        case 5:  Maske=32;  break;
        case 6:  Maske=64;  break;
        case 7:  Maske=128; break;
        /* dieser Fall sollte eigentlich nie eintreten: */
        default: Maske=0;   break;
    }
    M.elem[i/8]^=Maske;
}
